/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ArmarXObjects
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::ArmarXObjects

#define ARMARX_BOOST_TEST

#include <RobotAPI/Test.h>
#include "../ArmarXObjects.h"

#include "../ObjectID.h"

#include <iostream>


namespace armarx
{
    struct Fixture
    {
        std::vector<armarx::ObjectID> dcsStringConstructor
        {
            { "Data/Class/0" },
            { "Data/Class/1" },
            { "Data/Class/2" }
        };
        std::vector<armarx::ObjectID> dcsFromString
        {
            ObjectID::FromString("Data/Class/0"),
            ObjectID::FromString("Data/Class/1"),
            ObjectID::FromString("Data/Class/2")
        };
        std::vector<std::vector<armarx::ObjectID>*> dcs {
            &dcsStringConstructor,
            &dcsFromString
        };

        armarx::ObjectID dc0StringConstructor { "Data/Class/0" };
        armarx::ObjectID dc0FromString = ObjectID::FromString("Data/Class/0");
        std::vector<armarx::ObjectID*> dc0s {
            &dc0StringConstructor,
            &dc0FromString,
        };

        std::vector<armarx::ObjectID> otsStringConstructor
        {
            { "Other/Type/0" },
            { "Other/Type/1" },
            { "Other/Type/2" }
        };
        std::vector<armarx::ObjectID> otsFromString
        {
            ObjectID::FromString("Other/Type/0"),
            ObjectID::FromString("Other/Type/1"),
            ObjectID::FromString("Other/Type/2")
        };
        std::vector<std::vector<armarx::ObjectID>*> ots {
            &otsStringConstructor,
            &otsFromString
        };
    };
}

BOOST_FIXTURE_TEST_SUITE(ObjectIDTests, armarx::Fixture)

BOOST_AUTO_TEST_CASE(test_construction_from_string)
{
    for (const auto* dc : dcs)
    {
        for (const auto* ot : ots)
        {
            for (std::size_t i = 0; i < dc->size(); ++i)
            {
                BOOST_CHECK_EQUAL((*dc)[i].dataset(), "Data");
                BOOST_CHECK_EQUAL((*dc)[i].className(), "Class");
                BOOST_CHECK_EQUAL((*dc)[i].instanceName(), std::to_string(i));

                BOOST_CHECK_EQUAL((*ot)[i].dataset(), "Other");
                BOOST_CHECK_EQUAL((*ot)[i].className(), "Type");
                BOOST_CHECK_EQUAL((*ot)[i].instanceName(), std::to_string(i));
            }
        }
    }
}

BOOST_AUTO_TEST_CASE(test_equals_operator)
{
    for (const auto* dc : dcs)
    {
        for (const auto* ot : ots)
        {
            for (const auto* dc0 : dc0s)
            {
                for (std::size_t i = 0; i < (*dc).size(); ++i)
                {
                    BOOST_TEST_CONTEXT("i=" << i)
                    {
                        BOOST_CHECK_EQUAL((*dc)[i], (*dc)[i]);
                        BOOST_CHECK_NE((*dc)[i], (*ot)[i]);
                        if (i != 0)
                        {
                            BOOST_CHECK_NE((*dc)[i], *dc0);
                        }
                    }
                }
            }
        }
    }
}

BOOST_AUTO_TEST_CASE(test_less_than_operator)
{
    for (const auto* dc : dcs)
    {
        for (std::size_t i = 0; i < (*dc).size(); ++i)
        {
            for (std::size_t j = i; j < (*dc).size(); ++j)
            {
                BOOST_CHECK_LE((*dc)[i], (*dc)[j]);
                BOOST_CHECK_GE((*dc)[j], (*dc)[i]);
                if (i != j)
                {
                    BOOST_CHECK_LT((*dc)[i], (*dc)[j]);
                    BOOST_CHECK_GT((*dc)[j], (*dc)[i]);
                }
            }
        }
    }
}

BOOST_AUTO_TEST_CASE(test_equalClass)
{
    for (const auto* dc : dcs)
    {
        for (const auto* ot : ots)
        {
            for (std::size_t i = 0; i < (*dc).size(); ++i)
            {
                for (std::size_t j = 0; j < (*dc).size(); ++j)
                {
                    BOOST_CHECK((*dc)[i].equalClass((*dc)[j]));
                    BOOST_CHECK((*ot)[i].equalClass((*ot)[j]));

                    BOOST_CHECK(!(*dc)[i].equalClass((*ot)[j]));
                    BOOST_CHECK(!(*ot)[i].equalClass((*dc)[j]));
                }
            }
        }
    }
}

BOOST_AUTO_TEST_SUITE_END()
